"""
MIT License

Copyright (c) 2019-present Josh B

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Protocol

from ._consts import DEFAULT_SCHEMA
from .utils import MISSING, normalise_name

if TYPE_CHECKING:
    from asyncpg import Pool


__all__ = ("Object",)


class Object(Protocol):
    _schema: ClassVar[str]
    _local_name: ClassVar[str]
    _base_pool: ClassVar[Optional[Pool]] = None
    _overridden_name: ClassVar[Optional[str]] = None

    @classmethod
    @property
    def _pool(cls) -> Optional[Pool]:
        pool = None

        for cls in reversed(cls.mro()):
            pool = getattr(cls, "_base_pool", None) or pool

        return pool

    @classmethod
    @property
    def _name(cls) -> str:
        if cls._overridden_name is not None:
            return cls._overridden_name
        return f"{cls._schema}.{cls._local_name}"

    def __init_subclass__(
        cls,
        *,
        _name: str = MISSING,
        schema: str = MISSING,
    ) -> None:
        if schema is MISSING:
            schema = DEFAULT_SCHEMA

        if _name is MISSING:
            _name = normalise_name(cls.__name__)

        cls._schema = schema
        cls._local_name = _name
        super().__init_subclass__()

    @classmethod
    def set_pool(cls, pool: Pool) -> None:
        """
        A function which sets the default connection pool to use on this
        database onject and subclasses of it.

        Parameters
        ----------
        pool: :class:`asyncpg.Pool <asyncpg.pool.Pool>`
            The connection pool to use.
        """
        cls._base_pool = pool



"""
MIT License

Copyright (c) 2019-present Josh B

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THEdefault
SOFTWARE.
"""

from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, ClassVar

from ._consts import CUSTOM_TYPES
from ._creatable import Creatable
from ._custom_types import CustomType
from ._insertable import Insertable
from .utils import MISSING, not_creatable, optional_pool, optional_transaction, query_builder

if TYPE_CHECKING:
    from asyncpg import Connection, Record

    from ._column import Column


__all__ = ("Table",)


@not_creatable
class Table(Insertable, Creatable):
    """Base class for creating representations of SQL Database Tables.

    Attributes
    ----------
        _name: :class:`str`
            The name of the table.
        _schema: :class:`str`
            The tables schema.
        _columns: Iterable[:class:`~.Column`]
            The columns contained in the table.
        _columns_dict: Dict[:class:`str`, :class:`~.Column`]
            A mapping between a column's name and itself.
        _primary_keys: Iterable[:class:`~.Column`]
            The primary key columns of the table.
    """

    _type: ClassVar[str] = "TABLE"

    @classmethod
    def _query_exists(
        cls,
    ) -> str:
        return super()._query_exists("tables")

    @classmethod
    @query_builder
    def _query_create(cls, if_not_exists: bool) -> list[str]:
        builder = ["CREATE TABLE"]

        if if_not_exists:
            builder.append("IF NOT EXISTS")

        builder.append(cls._name)
        builder.append("(")

        unique_columns = []

        for column in cls._columns:
            builder.append(column._query())
            builder.append(",")

            if column.unique:
                unique_columns.append(column)

        if unique_columns:
            builder.append("UNIQUE (")
            for column in unique_columns:
                builder.append(column.name)
                builder.append(",")

            builder.pop(-1)
            builder.append(")")
            builder.append(",")

        if cls._primary_keys:
            builder.append("PRIMARY KEY (")
            for column in cls._primary_keys:
                builder.append(column.name)
                builder.append(",")

            builder.pop(-1)
            builder.append(")")
        else:
            builder.pop(-1)

        builder.append(")")

        return builder

    @classmethod
    @query_builder
    def _query_drop_column(cls, column: Column) -> list[str]:
        return ["ALTER TABLE", cls._name, "DROP COLUMN", column.name]

    @classmethod
    @query_builder
    def _query_add_column(cls, column: Column) -> list[str]:
        return ["ALTER TABLE", cls._name, "ADD COLUMN", column._query()]

    @classmethod
    @optional_pool
    async def create(
        cls,
        connection: Connection,
        *,
        if_not_exists: bool = True,
        create_schema: bool = True,
        automatic_migrations: bool = False,
    ) -> None:
        """|coro|

        Creates this database object.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        if_not_exists: :class:`bool`
            Sets whether creation should continue if the object already exists.
            Defaults to ``True``.
        create_schema: :class:`bool`
            Sets whether the database schema should also be created.
            Defaults to ``True``.
        automatic_migrations: :class:`bool`
            Sets whether migrations should be automatically run.
            Defaults to ``False``.
        """
        # create custom types if needed
        for column in cls._columns:
            if issubclass(column.sql_type, CustomType):
                if column.sql_type._name not in CUSTOM_TYPES:
                    await column.sql_type.create(connection, if_not_exists=if_not_exists)

        return await super().create(
            connection,
            if_not_exists=if_not_exists,
            create_schema=create_schema,
            automatic_migrations=automatic_migrations,
        )

    @classmethod
    @optional_pool
    async def drop_column(
        cls,
        connection: Connection,
        /,
        column: Column,
    ) -> None:
        """Drops a column from the table.

        Parameters
        ----------
            connection: :class:`~asyncpg.Connection`
                The connection to use.
            column: :class:`~.Column`
                The column to drop.
        """
        if getattr(column, "table", MISSING) is not cls:
            raise ValueError("Column does not belong to this table.")
        await connection.execute(cls._query_drop_column(column))
        del column.table

    @classmethod
    @optional_pool
    async def add_column(
        cls,
        connection: Connection,
        /,
        column: Column,
    ) -> None:
        """Adds a new column to the table.

        Parameters
        ----------
            connection: :class:`~asyncpg.Connection`
                The connection to use.
            column: :class:`~.Column`
                The column to add. This column must not be associated with
                another table.
        """
        if hasattr(column, "table"):
            raise ValueError(f"Column {column.name} already belongs to {column.table}")
        if column.name in cls._columns_dict:
            raise ValueError(f"Column {column.name} already exists.")
        cls._columns_dict[column.name] = column
        await connection.execute(cls._query_add_column(column))
        column.table = cls

    @classmethod
    @optional_pool
    async def migrate(
        cls,
        connection: Connection,
        /,
        columns: Iterable[Column],
        with_transaction: bool = True,
    ) -> None:
        """|coro|

        Migrates the table to the given columns.

        Parameters
        ----------
        connection: :class:`~asyncpg.Connection`
            The connection to use.
        columns: Iterable[:class:`~.Column`]
            The new list of columns for the table.
        with_transaction: :class:`bool`
            Sets whether the database should be wrapped in a transaction.
            Defaults to ``True``.
        """
        columns_dict = {column.name: column for column in columns}

        async with optional_transaction(connection, with_transaction):
            for column in cls._columns:
                if column.name not in columns_dict:
                    await cls.drop_column(connection, column)

            for column in columns_dict:
                if column not in cls._columns_dict:
                    await cls.add_column(connection, columns_dict[column])

    @classmethod
    @optional_pool
    async def migrate_to(
        cls,
        connection: Connection,
        /,
        table: type[Table],
        migration: Callable[[Record], dict[str, Any]] = MISSING,
        *,
        create_new_table: bool = False,
        drop_table: bool = False,
        with_transaction: bool = True,
    ) -> None:
        """|coro|

        Helper function for migrating data in a table to another.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        table: Type[:class:`.Table`]
            The new table to migrate to.
        migration: Callable[[:class:`asyncpg.Record`], Dict[:class:`str`, Any]]
            The function used to migrate data between tables.
            By default this passes the record unchanged.
        create_new_table: :class:`bool`
            Sets whether the table to migrate to should be created.
            Defaults to ``False``.
        drop_table: :class:`bool`
            Sets whether this table should be dropped after migrating.
            Defaults to ``False``.
        with_transaction: :class:`bool`
            Sets whether the database should be wrapped in a transaction.
            Defaults to ``True``.
        """
        if cls._name == table._name:
            if migration is not MISSING:
                raise ValueError("Custom migration functions are not supported for table alterations.")
            await cls.migrate(connection, (column._copy() for column in table._columns))
            return

        async with optional_transaction(connection, with_transaction):

            if migration is MISSING:
                migration = dict

                if create_new_table:
                    await table.create(connection)

                records = await cls.fetch(connection)
                await table.insert_many(connection, None, *(migration(record) for record in records))

                if drop_table:
                    await cls.drop(connection)



"""
MIT License

Copyright (c) 2019-present Josh B

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, TextIO, Union, overload

from ._column import Column
from ._consts import DEFAULT_SCHEMA, NOT_CREATABLE
from ._object import Object
from .utils import MISSING, optional_pool, optional_transaction, query_builder, write_to_file

if TYPE_CHECKING:
    from asyncpg import Connection


__all__ = ("Creatable",)


class Creatable(Object):
    _type: ClassVar[str] = MISSING

    # region: query generation

    @classmethod
    def _query_exists(
        cls,
    ) -> str:
        raise NotImplementedError()

    @classmethod
    def _query_create(
        cls,
        if_not_exists: bool,
    ) -> str:
        raise NotImplementedError()

    @classmethod
    @query_builder
    def _query_create_schema(
        cls,
        if_not_exists: bool,
    ) -> list[str]:
        builder = ["CREATE SCHEMA"]
        if if_not_exists:
            builder.append("IF NOT EXISTS")
        builder.append(cls._schema)
        return builder

    @classmethod
    @query_builder
    def _query_drop(
        cls,
        if_exists: bool,
        cascade: bool,
    ) -> list[str]:
        builder = ["DROP", cls._type]

        if if_exists:
            builder.append("IF EXISTS")

        builder.append(cls._name)

        if cascade:
            builder.append("CASCADE")

        return builder

    # endregion

    @classmethod
    def _find_schemas(cls) -> list[type[Creatable]]:
        schema_names: set[str] = set()
        schemas: list[type[Creatable]] = []

        for subcls in cls.__subclasses__():

            if subcls in NOT_CREATABLE:
                for schema in subcls._find_schemas():
                    if schema._schema not in schema_names:
                        schema_names.add(schema._schema)
                        schemas.append(schema)
            else:
                if subcls._schema not in schema_names:
                    schema_names.add(subcls._schema)
                    schemas.append(subcls)

        return schemas

    # region: public methods

    @classmethod
    @optional_pool
    async def exists(
        cls,
        connection: Connection,
    ) -> bool:
        """
        Check if the table exists.

        :param conn: The connection to use.
        :return: True if the table exists, False otherwise.
        """
        record = await connection.fetchrow(cls._query_exists(), cls._schema, cls._local_name)
        if record is None:
            return False
        (result,) = record
        return result

    @classmethod
    @optional_pool
    async def create(
        cls,
        connection: Connection,
        *,
        if_not_exists: bool = True,
        create_schema: bool = True,
        automatic_migrations: bool = False,
    ) -> None:
        """|coro|

        Creates this database object.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        if_not_exists: :class:`bool`
            Sets whether creation should continue if the object already exists.
            Defaults to ``True``.
        create_schema: :class:`bool`
            Sets whether the database schema should also be created.
            Defaults to ``True``.
        automatic_migrations: :class:`bool`
            Sets whether migrations should be automatically run.
            Defaults to ``False``.
        """
        if create_schema:
            await cls.create_schema(connection, if_not_exists=if_not_exists)

        if automatic_migrations:
            # this is a hack because >circular imports<
            from ._table import Table

            if issubclass(cls, Table) and await cls.exists(connection):
                new_columns = [column._copy() for column in cls._columns]
                old_columns: set[str] = {
                    record["column_name"]
                    for record in await connection.fetch(
                        "SELECT * FROM information_schema.columns WHERE table_schema = $1 AND table_name = $2;",
                        cls._schema,
                        cls._local_name,
                    )
                }

                cls._columns_dict = {column.name: column for column in cls._columns if column.name in old_columns}
                for column_name in old_columns:
                    if column_name not in cls._columns_dict:
                        column = Column.create(column_name, MISSING)
                        column.table = cls
                        cls._columns_dict[column_name] = column

                await cls.migrate(connection, new_columns)
                return

        query = cls._query_create(if_not_exists)
        await connection.execute(query)

    @classmethod
    @optional_pool
    async def create_schema(
        cls,
        connection: Connection,
        *,
        if_not_exists: bool = True,
    ) -> None:
        """|coro|

        Creates the schema this database object uses.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        if_not_exists: :class:`bool`
            Sets whether creation should continue if the schema already exists.
            Defaults to ``True``.
        """
        query = cls._query_create_schema(if_not_exists)
        await connection.execute(query)

    @classmethod
    @optional_pool
    async def create_all(
        cls,
        connection: Connection,
        /,
        *,
        if_not_exists: bool = True,
        create_schema: bool = True,
        automatic_migrations: bool = False,
        with_transaction: bool = True,
    ) -> None:
        """|coro|

        Creates all subclasses of this database object.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        if_not_exists: :class:`bool`
            Sets whether creation should continue if the object already exists.
            Defaults to ``True``.
        create_schema: :class:`bool`
            Sets whether the database schema should also be created.
            Defaults to ``True``.
        automatic_migrations: :class:`bool`
            Sets whether migrations should be automatically performed.
            Defaults to ``False``.
        with_transaction: :class:`bool`
            Sets whether the database should be wrapped in a transaction.
            Defaults to ``True``.
        """
        async with optional_transaction(connection, with_transaction):

            for subcls in cls.__subclasses__():
                if create_schema:
                    for schema in cls._find_schemas():
                        await schema.create_schema(connection, if_not_exists=if_not_exists)

                if subcls in NOT_CREATABLE:
                    await subcls.create_all(
                        connection,
                        if_not_exists=if_not_exists,
                        automatic_migrations=automatic_migrations,
                        with_transaction=False,
                    )
                else:
                    await subcls.create(
                        connection,
                        if_not_exists=if_not_exists,
                        automatic_migrations=automatic_migrations,
                    )

    @overload
    @classmethod
    def export(cls, *, if_not_exists: bool = ..., export_schema: bool = True, fp: None = ...) -> str:
        ...

    @overload
    @classmethod
    def export(cls, *, if_not_exists: bool = ..., export_schema: bool = True, fp: Union[str, TextIO] = ...) -> TextIO:
        ...

    @classmethod
    def export(
        cls, *, if_not_exists: bool = False, export_schema: bool = True, fp: Optional[Union[str, TextIO]] = None
    ) -> Union[TextIO, str]:
        """
        A function which exports this database object.

        Parameters
        ----------
        if_not_exists: :class:`bool`
            Sets whether the if_not_exists clause should be set on
            exported objects. Defaults to ``False``.
        export_schema: :class:`bool`
            Sets whether to additionally export the schema used by this object.
            Defaults to ``True``.
        fp: Optional[:class:`os.PathLike`, :class:`io.TextIOBase`]
            A file-like object opened in text mode and write mode.
            or a filename representing a file on disk to write to.

            .. note::
                If the file-like object passed is opened via :func:`open`
                ensure the object is in a text-writing mode such as ``"w"``.

        Returns
        -------
        Union[:class:`io.TextIOBase`, :class:`str`]
            The file-like object which was provided or a string containing the
            exported database object.
        """
        output = ""

        if export_schema and cls._schema != DEFAULT_SCHEMA:
            output += cls.export_schema(if_not_exists=if_not_exists)
            output += "\n\n"

        output += cls._query_create(if_not_exists)
        output += ";"

        if fp is None:
            return output
        return write_to_file(fp, output)

    @overload
    @classmethod
    def export_schema(cls, *, if_not_exists: bool = ..., fp: None = ...) -> str:
        ...

    @overload
    @classmethod
    def export_schema(cls, *, if_not_exists: bool = ..., fp: Union[str, TextIO] = ...) -> TextIO:
        ...

    @classmethod
    def export_schema(
        cls, *, if_not_exists: bool = False, fp: Optional[Union[str, TextIO]] = None
    ) -> Union[TextIO, str]:
        """|coro|

        A function which exports this database object's schema.

        Parameters
        ----------
        if_not_exists: :class:`bool`
            Sets whether the if_not_exists clause should be set on
            the exported database schema. Defaults to ``False``.
        fp: Optional[:class:`os.PathLike`, :class:`io.TextIOBase`]
            A file-like object opened in text mode and write mode.
            or a filename representing a file on disk to write to.

            .. note::
                If the file-like object passed is opened via :func:`open`
                ensure the object is in a text-writing mode such as ``"w"``.

        Returns
        -------
        Union[:class:`io.TextIOBase`, :class:`str`]
            The file-like object which was provided or a string containing the
            exported database schema.
        """
        output = cls._query_create_schema(if_not_exists)
        output += ";"

        if fp is None:
            return output
        return write_to_file(fp, output)

    @overload
    @classmethod
    def export_all(cls, *, if_not_exists: bool = ..., export_schema: bool = True, fp: None = ...) -> str:
        ...

    @overload
    @classmethod
    def export_all(
        cls, *, if_not_exists: bool = ..., export_schema: bool = True, fp: Union[str, TextIO] = ...
    ) -> TextIO:
        ...

    @classmethod
    def export_all(
        cls, *, if_not_exists: bool = False, export_schema: bool = True, fp: Optional[Union[str, TextIO]] = None
    ) -> Union[TextIO, str]:
        """
        A function which exports all database objects that subclass this object.

        Parameters
        ----------
        if_not_exists: :class:`bool`
            Sets whether the if_not_exists clause should be set on
            exported objects. Defaults to ``False``.
        export_schema: :class:`bool`
            Sets whether to additionally export any schemas used by these objects.
            Defaults to ``True``.
        fp: Optional[:class:`os.PathLike`, :class:`io.TextIOBase`]
            A file-like object opened in text mode and write mode.
            or a filename representing a file on disk to write to.

            .. note::
                If the file-like object passed is opened via :func:`open`
                ensure the object is in a text-writing mode such as ``"w"``.

        Returns
        -------
        Union[:class:`io.TextIOBase`, :class:`str`]
            The file-like object which was provided or a string containing the
            exported database objects.
        """
        output = ""

        if export_schema:
            for schema in cls._find_schemas():
                if schema._schema != DEFAULT_SCHEMA:
                    output += schema.export_schema(if_not_exists=if_not_exists)
                    output += "\n\n"

        for subcls in cls.__subclasses__():
            if subcls in NOT_CREATABLE:
                output += subcls.export_all(if_not_exists=if_not_exists, export_schema=False)
            else:
                output += subcls.export(if_not_exists=if_not_exists, export_schema=False)
                output += "\n\n"

        if fp is None:
            return output
        return write_to_file(fp, output)

    @classmethod
    @optional_pool
    async def drop(
        cls,
        connection: Connection,
        /,
        *,
        if_exists: bool = True,
        cascade: bool = False,
    ) -> None:
        """|coro|

        Drops this object from the database.

        Parameters
        ----------
        if_exists: :class:`bool`
            Sets whether dropping should not error if the object does not exist.
            Defaults to ``True``.
        cascade: :class:`bool`
            Sets whether dropping should cascade.
            Defaults to ``False``.
        """
        query = cls._query_drop(if_exists, cascade)
        await connection.execute(query)

    # endregion



"""
MIT License

Copyright (c) 2019-present Josh B

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import datetime
import json
from collections.abc import Callable
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Final,
    Literal,
    Optional,
    TextIO,
    TypeVar,
    Union,
    overload,
)

import asyncpg

from ._consts import CUSTOM_TYPES, DEFAULT_SCHEMA, POOLS
from .utils import DOCS_BUILDING, write_to_file, optional_transaction

if TYPE_CHECKING:
    from asyncpg import Connection, Pool

    from ._object import Object

__all__ = (
    "create_pool",
    "create_db",
    "export_db",
    "MaybeAcquire",
    "TYPE_CODECS",
    "TypeCodec",
    "OPTIONAL_CODECS",
)


T = TypeVar("T")
BE = TypeVar("BE", bound=BaseException)
OT = TypeVar("OT", bound="Object")


# Y2K_DT = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)
Y2K_EPOCH: Final[int] = 946684800000000


class TypeCodec(tuple[T]):
    """
    A NamedTuple defining a custom type codec.

    See :meth:`asyncpg.Connection.set_type_codec <asyncpg.connection.Connection.set_type_codec>` for more information.


    Parameters
    ----------
    format: Literal[``"text"``, ``"binary"``, ``"tuple"``]
        The type of decoder/encoder to use.

    encoder: collections.abc.Callable[``...``, Any]
        A callable which given a python object returns an encoded value.

    decoder: collections.abc.Callable[``...``, Any]
        A callable which given an encoded value returns a python object.
    """

    if TYPE_CHECKING:
        format: Literal["text", "binary", "tuple"]
        encoder: Callable[[T], Any]
        decoder: Callable[..., T]

    def __new__(
        cls: type[TypeCodec[T]],
        format: Literal["text", "binary", "tuple"],
        encoder: Callable[[T], Any],
        decoder: Callable[..., T],
    ) -> TypeCodec[T]:
        new_cls = super().__new__(cls, (format, encoder, decoder))  # type: ignore
        new_cls.format = format
        new_cls.encoder = encoder
        new_cls.decoder = decoder
        return new_cls


def _encode_datetime(value: datetime.datetime) -> tuple[int]:
    if value.tzinfo is None:
        value = value.astimezone(datetime.timezone.utc)
    return (round(value.timestamp() * 1_000_000) - Y2K_EPOCH,)


def _decode_timestamp(value: tuple[int]) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(round(value[0] + Y2K_EPOCH) / 1_000_000, datetime.timezone.utc)


TYPE_CODECS: dict[str, TypeCodec[Any]] = {
    "json": TypeCodec[dict]("text", json.dumps, json.loads),
    "jsonb": TypeCodec[dict]("text", json.dumps, json.loads),
}

OPTIONAL_CODECS: dict[str, TypeCodec[Any]] = {
    "timestamp": TypeCodec[datetime.datetime]("tuple", _encode_datetime, _decode_timestamp),
}

if DOCS_BUILDING and not TYPE_CHECKING:

    class TYPE_CODECS(dict[str, TypeCodec[Any]]):
        """
        A dictionary of pre-defined custom type-codecs.
        """

        @property
        def json(cls):
            """Automatically conveters a dictionary to and from json data to be stored in the database."""
            ...

        jsonb = json

    class OPTIONAL_CODECS(dict[str, TypeCodec[Any]]):
        """
        A dictionary of optional custom type-codecs.
        """

        @property
        def timestamp(cls):
            """Ensures all timestamps are timezone aware relative to UTC."""
            ...


async def create_pool(
    dsn: str,
    codecs: dict[str, TypeCodec[Any]] = {},
    *,
    set_as_default: bool = False,
    **kwargs: Any,
) -> Pool:
    r"""|coro|

    Creates the database connection pool.

    Parameters
    ----------
    dsn: :class:`str`
        A database connection string.
    codecs: Dict[:class:`str`, :class:`.TypeCodec`]
        A mapping of type to encoder and decoder for custom type codecs.
        A pre-defined set of codecs is provided in :class:`TYPE_CODECS` is used by default.
        As well as a set of :class:`OPTIONAL_CODECS` is provided.
    set_as_default: :class:`bool`
        Sets whether the pool should be used as the default connection pool for database operations.
        Defaults to ``False``.
    \*\*kwargs: Any
        Extra keyword arguments to pass to :func:`asyncpg.create_pool <asyncpg.pool.create_pool>`

    Returns
    -------
    :class:`asyncpg.Pool <asyncpg.pool.Pool>`
        The new pool which was created.
    """
    codecs = TYPE_CODECS | codecs

    async def init(connection: Connection) -> None:
        for type, codec in codecs.items():
            await connection.set_type_codec(
                type,
                schema="pg_catalog",
                encoder=codec.encoder,
                decoder=codec.decoder,
                format=codec.format,
            )

        for type in CUSTOM_TYPES.values():
            await type._set_codec(connection)

    pool = await asyncpg.create_pool(dsn, init=init, **kwargs)
    if pool is None:
        raise RuntimeError("Could not create pool.")
    POOLS.append(pool)

    if set_as_default:
        # this is a hack because >circular imports<
        from ._object import Object

        Object.set_pool(pool)

    return pool


async def create_db(
    connection: Connection,
    if_not_exists: bool = True,
    automatic_migrations: bool = False,
    with_transaction: bool = True,
) -> None:
    """|coro|

    A helper function to create all objects in the database.

    Parameters
    ----------
    connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
    if_not_exists: :class:`bool`
        Sets whether creation should continue if the object already exists.
        Defaults to ``True``.
    automatic_migrations: :class:`bool`
        Sets whether automatic migrations should be run.
        Defaults to ``False``.
    with_transaction: :class:`bool`
        Sets whether the database should be wrapped in a transaction.
        Defaults to ``True``.
    """
    # this is a hack because >circular imports<
    from ._creatable import Creatable
    from ._custom_types import CustomType
    from ._table import Table
    from ._view import View

    async with optional_transaction(connection, with_transaction):

        for schema in Creatable._find_schemas():
            if schema._schema != DEFAULT_SCHEMA:
                await schema.create(connection, if_not_exists=if_not_exists)

    await CustomType.create_all(
        connection,
        if_not_exists=if_not_exists,
        create_schema=False,
        automatic_migrations=automatic_migrations,
        with_transaction=False,
    )

    async with optional_transaction(connection, with_transaction):

        for type in (Table, View):
            await type.create_all(
                connection,
                if_not_exists=if_not_exists,
                create_schema=False,
                automatic_migrations=automatic_migrations,
                with_transaction=False,
            )


@overload
def export_db(*, if_not_exists: bool = ..., fp: None = ...) -> str:
    ...


@overload
def export_db(*, if_not_exists: bool = ..., fp: Union[str, TextIO] = ...) -> TextIO:
    ...


def export_db(*, if_not_exists: bool = False, fp: Optional[Union[str, TextIO]] = None) -> Union[TextIO, str]:
    """
    A helper function which exports all objects in the database.

    Parameters
    ----------
    if_not_exists: :class:`bool`
        Sets whether the if_not_exists clause should be set on
        exported objects. Defaults to ``False``.
    fp: Optional[:class:`os.PathLike`, :class:`io.TextIOBase`]
        A file-like object opened in text mode and write mode.
        or a filename representing a file on disk to write to.

        .. note::
            If the file-like object passed is opened via :func:`open`
            ensure the object is in a text-writing mode such as ``"w"``.

            If the file-like object passed is a path it will be opened in
            write-mode ``"w"``.

    Returns
    -------
    Union[:class:`io.TextIOBase`, :class:`str`]
        The file-like object which was provided or a string containing the
        exported database.
    """
    # this is a hack because >circular imports<
    from ._creatable import Creatable
    from ._custom_types import CustomType
    from ._table import Table
    from ._view import View

    output = ""

    for schema in Creatable._find_schemas():
        if schema._schema != DEFAULT_SCHEMA:
            output += schema.export_schema(if_not_exists=if_not_exists)
            output += "\n\n"

    for type in (CustomType, Table, View):
        output += type.export_all(if_not_exists=if_not_exists, export_schema=False)

    if fp is None:
        return output
    return write_to_file(fp, output)


class MaybeAcquire:
    """Async helper for acquiring a connection to the database.

    .. container:: operations

        .. describe:: async with x as c

            Yields a database connection to use,
            if a new one was created it will be closed on exit.

    Parameters
    ----------
        connection: Optional[:class:`asyncpg.Connection <asyncpg.connection.Connection>`]
            A database connection to use.
            If none is supplied a connection will be acquired from the pool.
        pool: Optional[:class:`asyncpg.Pool <asyncpg.pool.Pool>`]
            A connection pool to use.

    Attributes
    ----------
        connection: Optional[:class:`asyncpg.Connection <asyncpg.connection.Connection>`]
            The supplied database connection, if provided.
        pool: Optional[:class:`asyncpg.Pool <asyncpg.pool.Pool>`]
            The connection pool used to acquire new connections.
    """

    @overload
    def __init__(self, connection: Connection = ..., /, *, pool: Optional[Pool] = ...):
        ...

    @overload
    def __init__(self, connection: None = ..., /, *, pool: Pool = ...):
        ...

    def __init__(self, connection: Optional[Connection] = None, /, *, pool: Optional[Pool] = None):
        self.connection: Optional[Connection] = connection
        self.pool: Optional[Pool] = pool
        self._cleanup: bool = False

    async def __aenter__(self) -> Connection:
        if self.connection is None:
            self._cleanup = True
            self._connection = c = await self.pool.acquire()  # type: ignore
            return c
        return self.connection

    async def __aexit__(self, exc_type: type[BE], exc_val: BE, exc_tb: TracebackType) -> None:
        if self._cleanup:
            await self.pool.release(self._connection)  # type: ignore



"""
MIT License

Copyright (c) 2019-present Josh B

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, ClassVar, Literal

from ._column import BaseColumn, JoinColumn, OnClause
from ._consts import OPERATORS
from ._selectable import Selectable
from .utils import generate_alias, query_builder

JoinType = Literal["INNER", "LEFT", "RIGHT", "FULL OUTER"]

__all__ = ("Join",)


class Join(Selectable):
    """Base class for representations of SQL Database Tables.

    .. note::
        Subclasses of this class should never be created, instead
        join methods such as :meth:`Table.inner_join` should be used
        to create this model.

    Attributes
    ----------
        _name: :class:`str`
            The name of the table.
        _schema: :class:`str`
            The tables schema.
        _columns: Iterable[:class:`~.ViewColumn`]
            The columns contained in the table.
        _columns_dict: Dict[:class:`str`, :class:`~.ViewColumn`]
            A mapping between a column's name and itself.
        _query: :class:`str`
            The query used to create the view.
    """

    if TYPE_CHECKING:
        _columns: ClassVar[list[JoinColumn]]
        _columns_dict: ClassVar[dict[str, JoinColumn]]

    if TYPE_CHECKING:
        _type: ClassVar[JoinType]
        _alias: ClassVar[str]
        _selectables: tuple[type[Selectable], type[Selectable]]
        _aliases: dict[type[Selectable], str]
        _on: ClassVar[Iterable[OnClause]]

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        cls._set_columns()
        cls._overridden_name = cls._build_select()

    @classmethod
    def set_column(cls, source: BaseColumn) -> None:
        column = JoinColumn()
        column.name = source.name
        column.join = cls
        column.source = source._selectable

        cls._columns_dict[column.name] = column
        setattr(cls, column.name, column)

    @classmethod
    def _set_columns(cls) -> None:
        a, b = cls._selectables

        for column in a._columns:
            cls.set_column(column)

        for column in b._columns:
            if column.name in cls._columns_dict:
                continue
            cls.set_column(column)

    @classmethod
    @query_builder
    def _build_select(cls) -> list[str]:
        builder = ["( SELECT"]

        a, b = cls._selectables

        # Generate aliases
        if issubclass(a, Join):
            alias_a = a._alias
        else:
            alias_a = generate_alias()

        # Generate aliases
        if issubclass(b, Join):
            alias_b = b._alias
        else:
            alias_b = generate_alias()

        cls._aliases = {
            a: alias_a,
            b: alias_b,
        }

        for column in cls._columns:
            alias = cls._aliases[column.source]

            builder.append(f"{alias}.{column.name}")
            builder.append("AS")
            builder.append(column.name)
            builder.append(",")

        builder.pop(-1)

        builder.append("FROM")

        builder.append(a._name)
        if not issubclass(a, Join):
            builder.append("AS")
            builder.append(alias_a)

        builder.append(cls._type)
        builder.append("JOIN")

        builder.append(b._name)
        if not issubclass(b, Join):
            builder.append("AS")
            builder.append(alias_b)

        builder.append("ON")

        for clause in cls._on:
            if len(clause) == 2:
                col_a, col_b = clause  # type: ignore
                operator = "eq"
            else:
                col_a, col_b, operator = clause

            builder.append(f"{alias_a}.{col_a.name}")

            builder.append(OPERATORS[operator])

            builder.append(f"{alias_b}.{col_b.name}")

            builder.append("AND")

        builder.pop(-1)

        builder.append(") AS")

        builder.append(cls._alias)

        return builder
